{-# LANGUAGE TemplateHaskell #-}
{-# LANGUAGE NoMonomorphismRestriction#-}

module HAL.Client where

import Data.Proxy
import GHC.Generics
import Servant.API
import Servant.Client hiding (Response)
import Data.Text

import Data.Map
import Data.Aeson
import qualified Data.ByteString.Lazy as BSL
import qualified Codec.Binary.UTF8.String as UTF

import Control.Lens as L (makeLenses)

type HALAPI doc = Search doc
  :<|> Structure doc

type Search doc = "search"
  -- fl determine which fields will be returned it can be a list of fields or *
  :> QueryParam "fl" Text -- doc
  -- TODO: type this monster
  -- fq is to filter request
  :> QueryParams "fq" Text
  -- pretty much clear, (Asc || Desc) + field you want to sort by
  :> QueryParam "sort" SortField
  -- permit to start a the x result
  :> QueryParam "start" Int
  -- use rows to make the request only return the x number of result
  :> QueryParam "rows" Integer
  :> Get '[JSON] (Response doc)

type Structure doc = "ref" :> "structure"
  :> QueryParam "fl" Text
  :> QueryParam "fq" Text
  :> QueryParam "rows" Int
  :> Get '[JSON] (Response doc)

-- Get's argument type
data SortField = Asc Text | Desc Text
  deriving (Show)

instance ToHttpApiData SortField where
  toUrlPiece (Asc  t) = t <> " asc"
  toUrlPiece (Desc t) = t <> " desc"

asc :: Text -> Maybe SortField
asc = Just . Asc

desc :: Text -> Maybe SortField
desc = Just . Desc

-- Response type
data Response doc = Response
  { _numFound :: Integer
  , _start    :: Int
  , _docs     :: [doc]
  } deriving (Show, Generic)
L.makeLenses ''Response

instance FromJSON doc => FromJSON (Response doc) where
  parseJSON (Object o) = Response <$>
                         ((o .: "response") >>= (.: "numFound"))
                         <*> ((o .: "response") >>= (.: "start"))
                         <*> ((o .: "response") >>= (.: "docs"))

halAPI :: Proxy (HALAPI doc)
halAPI = Proxy

-- search should always have at least `docid` and `label_s` in his fl params
search :: (FromJSON doc, ToHttpApiData doc) =>
  Maybe Text -- fl
  -> [Text] -- fq
  -> Maybe SortField -- sort
  -> Maybe Int -- start
  -> Maybe Integer -- rows
  -> ClientM (Response doc)

structure :: (FromJSON doc, ToHttpApiData doc) =>
  Maybe Text
  -> Maybe Text -- fq
  -> Maybe Int -- rows
  -> ClientM (Response doc)

(search :<|> structure) = client halAPI
